/**
 * Copyright (C) 2013 Kai T. Ohlhus <k.ohlhus@gmail.com>
 *
 * This file is part of Octave.
 *
 * Octave is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * Octave is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Octave; see the file COPYING.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * This file implements the Incomplete Cholesky Factorization with zero fill.
 *
 * TODO: - Check if MIC(0) has been implemented correctly. Does not match with
 *         MATLAB at the moment.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "defun-dld.h"

/**
 * Performs the multiplication needed by the Cholesky factorization in the
 * complex case.
 *
 * @tparam data_t The type of the data operated on. In the complex case it is
 *   "Complex".
 * @param a First factor of multiplication.
 * @param b Second factor of multiplication.
 * @return a multiplied with the complex conjugate of b.
 */
template < typename data_t > inline data_t
ichol_mult_complex (data_t a, data_t b)
{
  b.imag (-std::imag (b));
  return a * b;
}

/**
 * Performs the multiplication needed by the Cholesky factorization in the
 * real case.
 *
 * @tparam data_t The type of the data operated on. In the real case it is
 *   "double".
 * @param a First factor of multiplication.
 * @param b Second factor of multiplication.
 * @return a multiplied b.
 */
template < typename data_t > inline data_t
ichol_mult_real (data_t a, data_t b)
{
  return a * b;
}

/**
 * Performs the standard Incomplete Cholesky Factorization [IC(0)] of the lower
 * triangular part of a symmetric positive definite sparse matrix A. A must be
 * split into its main diagonal in_diag and the non-diagonal sparse matrix
 * in_A.
 *
 * @tparam octave_sparse_matrix_t The type of the off diagonal sparse matrix.
 *   In the real case it is "SparseMatrix" and in the complex case
 *   "SparseComplexMatrix".
 * @tparam octave_column_vector_t The type of the off diagonal vector. In the
 *   real case it is "ColumnVector" and in the complex case
 *   "ComplexColumnVector".
 * @tparam data_t The type of the data operated on. In the real case it will
 *   be "double".
 * @tparam ichol_mult The multiplication operation needed by the Cholesky
 *   factorization. In the real case it is "ichol_mult_real" and in the
 *   complex case "ichol_mult_complex".
 *
 * @param in_A Non-diagonal non-zero entries of the lower triangular part of
 *   input matrix A.
 * @param in_diag The main diagonal of input matrix A.
 * @param michol Decides if the Modified IC(0) should be performed.
 *
 * @return on success a octave_value_list containing a sparse matrix and
 *   diagonal of the same size as the inputs with the decomposed input matrix
 *   A. Otherwise the row/column number of the zero diagonal.
 */
template < typename octave_sparse_matrix_t, typename octave_column_vector_t,
  typename data_t, data_t (*ichol_mult) (data_t, data_t) > octave_value_list
ichol0 (octave_sparse_matrix_t & in_A, octave_column_vector_t & in_diag,
	bool michol)
{
  octave_value_list retval;

  octave_idx_type n = in_A.cols ();
  octave_idx_type nnz = in_A.nnz ();

  // Allocate memory for the decomposed matrix
  octave_column_vector_t result_diag (n);
  octave_sparse_matrix_t result_A (n, n, nnz);

  // Define convenience pointers to the data structures
  data_t *diag = result_diag.fortran_vec ();
  data_t *off_data = result_A.data ();
  octave_idx_type *cidx = result_A.cidx ();
  octave_idx_type *ridx = result_A.ridx ();

  // Allocate working memory
  OCTAVE_LOCAL_BUFFER (data_t, ta, n);
  OCTAVE_LOCAL_BUFFER (octave_idx_type, ifirst, n);
  OCTAVE_LOCAL_BUFFER (octave_idx_type, list, n);

  // Copy input data to work and result data structures
  for (octave_idx_type i = 0; i < n; i++)
    diag[i] = in_diag.elem (i);
  for (octave_idx_type i = 0; i < nnz; i++)
    off_data[i] = in_A.data (i);
  for (octave_idx_type i = 0; i < (n + 1); i++)
    cidx[i] = in_A.cidx (i);
  for (octave_idx_type i = 0; i < nnz; i++)
    ridx[i] = in_A.ridx (i);

  // Initialize data structures
  for (octave_idx_type i = 0; i < n; i++)
    {
      ifirst[i] = -1;
      list[i] = -1;
    }

  // loop over all columns
  for (octave_idx_type k = 0; k < n; k++)
    {
      // load column k into ta and set ifirst
      octave_idx_type col_start_idx = cidx[k];
      octave_idx_type col_end_idx = cidx[k + 1] - 1;
      for (octave_idx_type j = col_start_idx; j <= col_end_idx; j++)
	{
	  octave_idx_type row = ridx[j];
	  ta[row] = off_data[j];
	  ifirst[row] = 0;
	}

      // make sure the diagonal of k is okay and then take the sqrt
      if (diag[k] <= 0.0)
	{
	  error ("ichol0: zero diagonal found.");
	  retval (0) = octave_value (k);
	  return retval;
	}
      diag[k] = sqrt (diag[k]);

      // update column k using the previous columns
      octave_idx_type j = list[k];
      data_t dropsum = 0.0;
      while (j != -1)
	{
	  octave_idx_type col_start = ifirst[j];
	  octave_idx_type col_end = cidx[j + 1] - 1;
	  data_t lval = off_data[col_start];
	  col_start++;
	  if (col_start < col_end)
	    {
	      ifirst[j] = col_start;
	      octave_idx_type iptr = j;
	      j = list[j];
	      list[iptr] = list[ridx[col_start]];
	      list[ridx[col_start]] = iptr;
	    }
	  else
	    j = list[j];
	  for (octave_idx_type i = col_start; i <= col_end; i++)
	    {
	      octave_idx_type row = ridx[i];
	      if (ifirst[row] != -1)
		ta[row] -= ichol_mult (lval, off_data[i]);
	      else
		dropsum -= ichol_mult (lval, off_data[i]);
	    }
	}

      // ifist and list keep track of where in column k we are
      if (col_start_idx < col_end_idx)
	{
	  octave_idx_type iptr = ridx[col_start_idx];
	  list[k] = list[iptr];
	  list[iptr] = k;
	  ifirst[k] = col_start_idx;
	}

      // compensate diagonal MIC(0)
      if (michol)
	{
	  diag[k] += dropsum;
	  if (diag[k] <= 0.0)
	    {
	      error ("ichol0: zero diagonal found.");
	      retval (0) = octave_value (k);
	      return retval;
	    }
	}

      // update remaining diagonals using column k
      data_t lval = diag[k];
      for (j = col_start_idx; j <= col_end_idx; j++)
	{
	  int row = ridx[j];
	  data_t t = ta[row];
	  ifirst[row] = -1;
	  t /= lval;
	  diag[row] -= ichol_mult (t, t);
	  off_data[j] = t;
	}
    }

  retval (0) = octave_value (result_A);
  retval (1) = octave_value (result_diag);

  return retval;
}

DEFUN_DLD (ichol0, args, nargout, "-*- texinfo -*-\n\
@deftypefn  {Loadable Function} {[@var{L_off}, @var{D2}] =} ichol0 (@var{A_off}, @var{D1}, @var{michol})\n\
\n\
Computes the no fill Incomplete Cholesky factorization [IC(0)] of a \
symmetric positive definite matrix \
@code{@var{A} = @var{A_off} + @var{D1} + @var{A_off}^T}.\n\
\n\
@code{[@var{L_off}, @var{D2}] = ichol0 (@var{A_off}, @var{D1}, @var{michol})} \
computes the IC(0) of @var{A}, such that @code{@var{L} * @var{L}^T} with \
@code{@var{L} = @var{L_off} + @var{D2}} is an approximation of the square \
sparse symmetric matrix @var{A}. The parameter @var{michol} decides whether \
the Modified IC(0) should be performed. This compensates the main diagonal of \
@var{L}, such that @code{@var{A} * @var{e} = @var{L} * @var{L}^T * @var{e}} \
with @code{@var{e} = ones (size (@var{A}, 2), 1))} holds. \n\
\n\
For more information about the algorithms themselves see:\n\
\n\
[1] Jones, Mark T. and Plassmann, Paul E.: An Improved Incomplete Cholesky \
Factorization (1992)\n\
\n\
[2] Jones, Mark T. and Plassmann, Paul E.: Algorithm 740: Fortran subroutines \
to compute improved incomplete Cholesky factorizations (1995)\n\
\n\
This implementation is based upon [2].\n\
@seealso{ichol, ichol0jp, icholt, chol, ilu}\n\
@end deftypefn")
{
  octave_value_list retval;
  int nargin = args.length ();
  if ((nargin != 3) || nargout != 2)
    {
      print_usage ();
      return retval;
    }

  if (args (0).is_empty () || !args (0).is_matrix_type ()
      || !args (0).is_sparse_type ())
    {
      error ("ichol0: first parameter must be a sparse matrix.");
      return retval;
    }

  if (args (1).is_empty ())
    {
      error ("ichol0: second parameter must be vector with elements > 0.");
      return retval;
    }

  bool michol = false;
  if (args (2).is_empty () || !args (2).is_bool_type ())
    {
      error ("ichol0: third parameter must be a boolean.");
      return retval;
    }
  else
    michol = args (2).bool_value ();

  if (!error_state)
    {
      if (args (0).is_complex_type ())
	{
	  // Prepare input matrix complex case
	  SparseComplexMatrix A = args (0).sparse_complex_matrix_value ();
	  ComplexColumnVector D = args (1).complex_column_vector_value ();
	  retval =
	    ichol0 < SparseComplexMatrix, ComplexColumnVector, Complex,
	    ichol_mult_complex > (A, D, michol);
	}
      else
	{
	  // Prepare input matrix real case
	  SparseMatrix A = args (0).sparse_matrix_value ();
	  ColumnVector D = args (1).column_vector_value ();
	  retval =
	    ichol0 < SparseMatrix, ColumnVector, double,
	    ichol_mult_real > (A, D, michol);
	}
    }

  return retval;
}

/*
%!shared A_1, D_1, A_off_1, A_2, D_2, A_off_2, A_3, D_3, A_off_3, A_4, D_4, A_off_4, A_5, D_5, A_off_5
%! A_1 = [ 0.37, -0.05,  -0.05,  -0.07;
%!        -0.05,  0.116,  0.0,   -0.05;
%!        -0.05,  0.0,    0.116, -0.05;
%!        -0.07, -0.05,  -0.05,   0.202];
%! A_1 = sparse(A_1);
%! D_1 = diag (A_1);
%! A_off_1 = (tril (A_1 - diag (D_1)));
%! D_1 = full (D_1);
%!
%! A_2 = gallery ('poisson', 30);
%! D_2 = diag (A_2);
%! A_off_2 = (tril (A_2 - diag (D_2)));
%! D_2 = full (D_2);
%!
%! A_3 = gallery ('tridiag', 50);
%! D_3 = diag (A_3);
%! A_off_3 = (tril (A_3 - diag (D_3)));
%! D_3 = full (D_3);
%!
%! nx = 400; ny = 200;
%! hx = 1 / (nx + 1); hy = 1 / (ny + 1);
%! Dxx = spdiags ([ones(nx, 1), -2 * ones(nx, 1), ones(nx, 1)], [-1 0 1 ], nx, nx) / (hx ^ 2);
%! Dyy = spdiags ([ones(ny, 1), -2 * ones(ny, 1), ones(ny, 1)], [-1 0 1 ], ny, ny) / (hy ^ 2);
%! A_4 = -kron (Dxx, speye (ny)) - kron (speye (nx), Dyy);
%! D_4 = diag (A_4);
%! A_off_4 = (tril (A_4 - diag (D_4)));
%! D_4 = full (D_4);
%!
%! A_5 = [ 0.37, -0.05,          -0.05,  -0.07;
%!        -0.05,  0.116,          0.0,   -0.05 + 0.05i;
%!        -0.05,  0.0,            0.116, -0.05;
%!        -0.07, -0.05 - 0.05i,  -0.05,   0.202];
%! A_5 = sparse(A_5);
%! D_5 = diag (A_5);
%! A_off_5 = (tril (A_5 - diag (D_5)));
%! D_5 = full (D_5);
%!
%!test
%!error ichol0 ([]);
%!error ichol0 ([],[]);
%!error ichol0 ([],[],[]);
%!error [~,~] = ichol0 ([],[],[]);
%!error [L] = ichol0 ([],[],[]);
%!error [L, D] = ichol0 ([],[],[]);
%!error [L] = ichol0 (A_1);
%!error [L] = ichol0 (A_off_1, D_1);
%!error [L] = ichol0 (A_off_1, D_1, false);
%!error [L, D] = ichol0 ([], 4, false);
%!error [L, D] = ichol0 (A_off_1, [], false);
%!error [L, D] = ichol0 (A_off_1, D_1, []);
%!error [L] = ichol0 (A_off_1, D_1, false);
%!error [L, D, E] = ichol0 (A_off_1, D_1, false);
%!error [L, D] = ichol0 (A_off_1, D_1, false, A_off_1);
%!error ichol0 (sparse (0), 0, false);
%!error ichol0 (sparse (0), -0, false);
%!error ichol0 (sparse (0), 1, false);
%!error ichol0 (sparse (0), -1, false);
%!error ichol0 (sparse (0), i, false);
%!error ichol0 (sparse (0), -i, false);
%!error ichol0 (sparse (0), 1 + 1i, false);
%!error ichol0 (sparse (0), 1 - 1i, false);
%!error ichol0 (sparse (0), sparse (0), false);
%!error ichol0 (sparse (0), sparse (-0), false);
%!error ichol0 (sparse (0), sparse (-1), false);
%!error ichol0 (sparse (0), sparse (-1), false);
%!
%!test
%! [L, D] = ichol0 (sparse (0), sparse (1), false);
%! assert (L + sparse (diag (D)), sparse (1));
%! [L, D] = ichol0 (sparse (0), sparse (4), false);
%! assert (L + sparse (diag (D)), sparse (2));
%! [L, D] = ichol0 (sparse (0), 1 + 1i, false);
%! assert (L + sparse (diag (D)), sparse (1));
%! [L, D] = ichol0 (sparse (0), 1 - 1i, false);
%! assert (L + sparse (diag (D)), sparse (1));
%! [L, D] = ichol0 (sparse (0), 2 + 2i, false);
%! assert (L + sparse (diag (D)), sparse (sqrt (2)));
%!
%!test
%! [L, D] = ichol0 (A_off_1, D_1, false);
%! L = L + sparse (diag (D));
%! assert (norm (A_1 - L*L', 'fro') / norm (A_1, 'fro'), 1e-2, 1e-2);
%! [L, D] = ichol0 (A_off_1, D_1, true);
%! L = L + sparse (diag (D));
%! assert (norm (A_1 - L*L', 'fro') / norm (A_1, 'fro'), 2e-2, 1e-2);
%!
%!test
%! [L, D] = ichol0 (A_off_2, D_2, false);
%! L = L + sparse (diag (D));
%! assert (norm (A_2 - L*L', 'fro') / norm (A_2, 'fro'), 1e-1, 1e-1)
%!error [L, D] = ichol0 (A_off_2, D_2, true);
%!
%!test
%! [L, D] = ichol0 (A_off_3, D_3, false);
%! L = L + sparse (diag (D));
%! assert (norm (A_3 - L*L', 'fro') / norm (A_3, 'fro'), eps, eps);
%! [L, D] = ichol0 (A_off_3, D_3, true);
%! L = L + sparse (diag (D));
%! assert (norm (A_3 - L*L', 'fro') / norm (A_3, 'fro'), eps, eps);
%!
%!test
%! [L, D] = ichol0 (A_off_4, D_4, false);
%! L = L + sparse (diag (D));
%! assert (norm (A_4 - L*L', 'fro') / norm (A_4, 'fro'), 1e-1, 1e-1);
%!error [L, D] = ichol0 (A_off_4, D_4, true);
%!
%!test
%! [L, D] = ichol0 (A_off_5, D_5, false);
%! L = L + sparse (diag (D));
%! assert (norm (A_5 - L*L', 'fro') / norm (A_5, 'fro'), 1e-2, 1e-2);
%! [L, D] = ichol0 (A_off_5, D_5, true);
%! L = L + sparse (diag (D));
%! assert (norm (A_5 - L*L', 'fro') / norm (A_5, 'fro'), 2e-2, 1e-2);
*/

