/**
 * Copyright (C) 2013 Kai T. Ohlhus <k.ohlhus@gmail.com>
 *
 * This file is part of Octave.
 *
 * Octave is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * Octave is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Octave; see the file COPYING.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * This file implements the Incomplete Cholesky Factorization with threshold.
 *
 * TODO: - Remove Fortran indexing
 *       - Improve MIC
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "defun-dld.h"

/**
 * Performs the Incomplete Cholesky Factorization with threshold [ICT] of the
 * lower triangular part of a symmetric positive definite sparse matrix A.
 *
 * @tparam octave_sparse_matrix_t The type of the off diagonal sparse matrix.
 *   In the real case it is "SparseMatrix" and in the complex case
 *   "SparseComplexMatrix".
 * @tparam data_t The type of the data operated on. In the real case it will
 *   be "double".
 *
 * @param in_A Non-diagonal non-zero entries of the lower triangular part of
 *   input matrix A.
 * @param droptol Non-negative scalar used as a drop tolerance when performing
 *   ICT. Elements which are smaller in magnitude than a local drop tolerance
 *   are dropped from the resulting factor.
 * @param nl Decides how much working memory for the resulting matrix will
 *   be allocated. Suggested value: 2*nnz(A).
 * @param maxk Decides how much additional working memory will be allocated.
 *   Suggested value: 20 or n + 1.
 * @param michol Decides if the Modified IC(0) should be performed.
 *
 * @return on success a octave_value_list containing a sparse matrix of the same
 *   dimension as the inputs with the decomposed input matrix A. Otherwise the
 *   row/column number of the zero diagonal.
 */
template < typename octave_sparse_matrix_t,
  typename data_t > octave_value_list
icholt (octave_sparse_matrix_t in_A, data_t droptol,
	octave_idx_type nnz_guess, octave_idx_type maxk, bool michol)
{
  octave_value_list retval;

  octave_idx_type n = in_A.cols ();

  // Allocate workig arrays for output matrix
  OCTAVE_LOCAL_BUFFER (data_t, data, nnz_guess + 1);
  OCTAVE_LOCAL_BUFFER (octave_idx_type, ridx, nnz_guess + 1);
  OCTAVE_LOCAL_BUFFER (octave_idx_type, cidx, n + 2);
  OCTAVE_LOCAL_BUFFER (octave_idx_type, link, nnz_guess + 1);

  // Allocate workig arrays
  OCTAVE_LOCAL_BUFFER (octave_idx_type, icol, maxk + 1);
  OCTAVE_LOCAL_BUFFER (octave_idx_type, kpoint, maxk + 1);
  OCTAVE_LOCAL_BUFFER (data_t, diag, n + 1);

  // Copy diagonal into data and set link
  for (octave_idx_type i = 1; i <= n; i++)
    {
      diag[i] = in_A.data (in_A.cidx (i - 1));
      link[nnz_guess - n + i] = 0;
    }

  octave_idx_type diag_idx = 1;
  // start of column i
  for (octave_idx_type i = 1; i <= n; i++)
    {
      octave_idx_type column_start_idx = in_A.cidx (i - 1) + 1;
      octave_idx_type extra_elements = 0;
      octave_idx_type insert_idx = diag_idx;
      cidx[i] = diag_idx;
      octave_idx_type lpoint = link[nnz_guess - n + i];

      // iterate through extra elements
      while (lpoint != 0)
	{
	  extra_elements++;
	  icol[extra_elements] = lpoint;
	  if (i == ridx[lpoint + 1])
	    kpoint[extra_elements] = 1;
	  else
	    kpoint[extra_elements] = lpoint + 1;
	  if (extra_elements >= maxk)
	    {
	      error ("icholt: maxk was not set large enough.");
	      return retval;
	    }
	  lpoint = link[lpoint];
	}

      // check for correct diagonal in column i
      if (diag[i] <= 0.0)
	{
	  error
	    ("icholt: Pivot error. Check if matrix is symmetric positive definite.");
	  retval (0) = octave_value (i);
	  return retval;
	}

      octave_idx_type nextj = std::numeric_limits < octave_idx_type >::max ();
      // for each row j
      for (octave_idx_type j = i;
	   j < std::numeric_limits < octave_idx_type >::max (); j = nextj)
	{
	  octave_idx_type row_start_idx =
	    in_A.ridx (column_start_idx - 1) + 1;
	  data_t x = in_A.data (column_start_idx - 1);

	  nextj = std::numeric_limits < octave_idx_type >::max ();

	  // if missaligned advance j to starting row index of current column
	  if (j < row_start_idx)
	    nextj = row_start_idx;
	  if (j == row_start_idx)
	    {
	      // if current column contains more elements
	      if ((column_start_idx - in_A.cidx (i)) < 0)	// with Fortran: cidx[i + 1] - 1
		{
		  column_start_idx++;
		  nextj = in_A.ridx (column_start_idx - 1) + 1;
		}
	    }
	  else
	    x = 0.0;

	  // treat extra elements
	  for (octave_idx_type k = 1; k <= extra_elements; k++)
	    {
	      if (j == ridx[kpoint[k]])
		{
		  x -= data[kpoint[k]] * data[icol[k]];
		  kpoint[k]++;
		}
	      if ((j < ridx[kpoint[k]]) && (ridx[kpoint[k]] < nextj))
		nextj = ridx[kpoint[k]];
	    }

	  // insert element, if rejection criterion is not met
	  if ((x * x - droptol * droptol * diag[i] * diag[j]) < 0.0)
	    {
	      if (michol)
		{
		  data_t y = std::sqrt (diag[j] / diag[i]);
		  diag[i] += std::abs (x) * y;
		  diag[j] += std::abs (x) / y;
		}
	    }
	  else
	    {
	      data[insert_idx] = x;
	      ridx[insert_idx] = j;
	      link[insert_idx] = link[nnz_guess - n + j];
	      link[nnz_guess - n + j] = insert_idx;
	      insert_idx++;
	      if (insert_idx > (nnz_guess - n + i))
		{
		  error ("icholt: Not sufficient storage for output matrix.");
		  return retval;
		}
	    }
	}

      diag[i] = std::sqrt (diag[i]);
      data[diag_idx] = diag[i];

      if ((diag_idx + 1) <= (insert_idx - 1))
	for (octave_idx_type j = diag_idx + 1; j <= insert_idx - 1; j++)
	  {
	    // divide whole column by diagonal
	    data[j] /= diag[i];
	    // update lower diagonals
	    diag[ridx[j]] -= data[j] * data[j];
	  }
      diag_idx = insert_idx;
      cidx[n + 1] = insert_idx;	// write new nnz
    }

  octave_idx_type nnz = link[nnz_guess];
  SparseMatrix L (n, n, nnz);
  for (octave_idx_type i = 1; i <= nnz; i++)
    L.data (i - 1) = data[i];
  for (octave_idx_type i = 1; i <= nnz; i++)
    L.ridx (i - 1) = ridx[i] - 1;
  for (octave_idx_type i = 1; i <= (n + 1); i++)
    L.cidx (i - 1) = cidx[i] - 1;

  retval (0) = octave_value (L);
  return retval;
}


DEFUN_DLD (icholt, args, nargout, "-*- texinfo -*-\n\
@deftypefn  {Loadable Function} {[@var{L}] =} icholt (@var{A}, @var{droptol}, @var{michol})\n\
\n\
Computes the thresholded Incomplete Cholesky factorization [ICT] of a \
symmetric positive definite matrix \
@code{@var{A} = @var{A_off} + @var{D} + @var{A_off}^T}. The input @var{A} is \
only the lower triangle part @code{@var{A} = @var{A_off} + @var{D}}. \n\
\n\
@code{[@var{L}] = icholt (@var{A}, @var{droptol}, @var{michol})} \
computes the ICT of @var{A}, such that @code{@var{L} * @var{L}^T} is an \
approximation of the square sparse symmetric matrix @var{A}. @var{droptol} is \
a non-negative scalar used as a drop tolerance when performing ICT. Elements \
which are smaller in magnitude than a local drop tolerance are dropped from \
the resulting factor @var{L}. The parameter @var{michol} decides whether \
the Modified IC(0) should be performed. This compensates the main diagonal of \
@var{L}, such that @code{@var{A} * @var{e} = @var{L} * @var{L}^T * @var{e}} \
with @code{@var{e} = ones (size (@var{A}, 2), 1))} holds. \n\
\n\
For more information about the algorithms themselves see:\n\
\n\
[1] AJIZ, M. A. and JENNINGS, A.: A Robust Incomplete Choleski-Conjugate \
Gradient Algorithm (1984)\n\
\n\
This implementation is based upon [1].\n\
@seealso{ichol, ichol0, ichol0jp, chol, ilu}\n\
@end deftypefn")
{
  octave_value_list retval;
  int nargin = args.length ();
  if ((nargin != 3) || nargout != 1)
    {
      print_usage ();
      return retval;
    }

  if (args (0).is_empty () || !args (0).is_matrix_type ()
      || !args (0).is_sparse_type ())
    {
      error ("icholt: first parameter must be a sparse matrix.");
      return retval;
    }

  double droptol = 1e-4;
  if (args (1).is_empty () || !args (1).is_scalar_type ()
      || ((droptol = args (1).double_value ()) < 0))
    {
      error ("icholt: second parameter must be a positive real value.");
      return retval;
    }

  bool michol = false;
  if (args (2).is_empty () || !args (2).is_bool_type ())
    {
      error ("icholt: third parameter must be a boolean.");
      return retval;
    }
  else
    michol = args (2).bool_value ();

  if (!error_state)
    {
      if (args (0).is_complex_type ())
	error ("icholt: cannot handle complex input.");
      else
	{
	  // Prepare input matrix real case
	  SparseMatrix A = args (0).sparse_matrix_value ();
	  // TODO: allow manual tuning of these parameters
	  octave_idx_type nl = 10 * A.nnz ();
	  octave_idx_type maxk = A.cols () + 1;	// suggested 20
	  retval =
	    icholt < SparseMatrix, double >(A, droptol, nl, maxk, michol);
	}
    }

  return retval;
}

/*
%!shared A_1, A_1_in, A_2, A_2_in, A_3, A_3_in, A_4, A_4_in, A_5, A_5_in
%! A_1 = [ 0.37, -0.05,  -0.05,  -0.07;
%!        -0.05,  0.116,  0.0,   -0.05;
%!        -0.05,  0.0,    0.116, -0.05;
%!        -0.07, -0.05,  -0.05,   0.202];
%! A_1_in = sparse(tril (A_1));
%!
%! A_2 = gallery ('poisson', 30);
%! A_2_in = sparse(tril (A_2));
%!
%! A_3 = gallery ('tridiag', 50);
%! A_3_in = sparse(tril (A_3));
%!
%! nx = 400; ny = 200;
%! hx = 1 / (nx + 1); hy = 1 / (ny + 1);
%! Dxx = spdiags ([ones(nx, 1), -2 * ones(nx, 1), ones(nx, 1)], [-1 0 1 ], nx, nx) / (hx ^ 2);
%! Dyy = spdiags ([ones(ny, 1), -2 * ones(ny, 1), ones(ny, 1)], [-1 0 1 ], ny, ny) / (hy ^ 2);
%! A_4 = -kron (Dxx, speye (ny)) - kron (speye (nx), Dyy);
%! A_4_in = sparse(tril (A_4));
%!
%! A_5 = [ 0.37, -0.05,          -0.05,  -0.07;
%!        -0.05,  0.116,          0.0,   -0.05 + 0.05i;
%!        -0.05,  0.0,            0.116, -0.05;
%!        -0.07, -0.05 - 0.05i,  -0.05,   0.202];
%! A_5 = sparse(A_5);
%! A_5_in = sparse(tril (A_5));
%!
%!test
%!error icholt ([]);
%!error icholt ([],[]);
%!error icholt ([],[],[]);
%!error [~] = icholt ([],[],[]);
%!error [L] = icholt ([],[],[]);
%!error [L] = icholt (A_1_in);
%!error [L] = icholt (A_1_in, 1e-4);
%!error [L] = icholt ([], 1e-4, false)
%!error [L] = icholt (A_1_in, [], false)
%!error [L] = icholt (A_1_in, 1e-4, [])
%!error [L, E] = icholt (A_1_in, 1e-4, false)
%!error [L] = icholt (A_1_in, 1e-4, false, A_1)
%!error icholt (sparse (0), 1e-4, false);
%!error icholt (sparse (-0), 1e-4, false);
%!error icholt (sparse (1), 1e-4, false);
%!error icholt (sparse (-1), 1e-4, false);
%!error icholt (sparse (i), 1e-4, false);
%!error icholt (sparse (-i), 1e-4, false);
%!error icholt (sparse (1 + 1i), 1e-4, false);
%!error icholt (sparse (1 - 1i), 1e-4, false);
%!
%!test
%! [L] = icholt (sparse (1), 1e-4, false);
%! assert (L, sparse (1));
%! [L] = icholt (sparse (4), 1e-4, false);
%! assert (L, sparse (2));
%!
%!test
%! [L] = icholt (A_1_in, 1e-4, false);
%! assert (norm (A_1 - L*L', 'fro') / norm (A_1, 'fro'), eps, eps)
%! [L] = icholt (A_1_in, 1e-4, true);
%! assert (norm (A_1 - L*L', 'fro') / norm (A_1, 'fro'), eps, eps)
%!
%!test
%! [L] = icholt (A_2_in, 1e-4, false);
%! assert (norm (A_2 - L*L', 'fro') / norm (A_2, 'fro'), 1e-4, 1e-4)
%! [L] = icholt (A_2_in, 1e-4, true);
%! assert (norm (A_2 - L*L', 'fro') / norm (A_2, 'fro'), 2e-4, 1e-4)
%!
%!test
%! [L] = icholt (A_3_in, 1e-4, false);
%! assert (norm (A_3 - L*L', 'fro') / norm (A_3, 'fro'), eps, eps)
%! [L] = icholt (A_3_in, 1e-4, true);
%! assert (norm (A_3 - L*L', 'fro') / norm (A_3, 'fro'), eps, eps)
%!
%!test
%! [L] = icholt (A_4_in, 1e-4, false);
%! assert (norm (A_4 - L*L', 'fro') / norm (A_4, 'fro'), 1e-4, 1e-4)
%! [L] = icholt (A_4_in, 1e-4, true);
%! assert (norm (A_4 - L*L', 'fro') / norm (A_4, 'fro'), 3e-4, 1e-4)
%!
%!test
%!error [L] = icholt (A_5_in, 1e-4, false);
%!error [L] = icholt (A_5_in, 1e-4, true);
*/

