/**
 * Copyright (C) 2013 Kai T. Ohlhus <k.ohlhus@gmail.com>
 *
 * This file is part of Octave.
 *
 * Octave is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * Octave is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Octave; see the file COPYING.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * This is the octave interface to itsol. For more information see
 * http://www-users.cs.umn.edu/~saad/software/ITSOL/
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "defun-dld.h"

// Auxiliary ITSOL libraries
#include "itsol_util.h"

/**
 * Type generic ILUT implementation that performs the same procedure for real
 * and complex input and calls ITSOL or ZITSOL library functions accordingly.
 *
 * @tparam octave_sparse_matrix_t The type of the Octave input matrix. Intended
 *   values for ITSOL (SparseMatrix) and for ZITSOL (SparseComplexMatrix)
 * @tparam octave_column_vector_t The type for the diagonal vector that will be
 *   returned to Octave. Intended values for ITSOL (ColumnVector) and for ZITSOL
 *   (ComplexColumnVector)
 * @tparam c_data_t Data type used in ITSOL (double) and ZITSOL
 *   (complex_conversion_t)
 * @tparam cpp_data_t Data type used in Octave. Indended values for ITSOL
 *   (double) and for ZITSOL (Complex)
 * @tparam z_itsol_sparse_matrix_t Sparse matrix storage type used in ITSOL
 *   (itsol_sparse_matrix_t) and ZITSOL (zitsol_sparse_matrix_t)
 * @tparam z_itsol_decomposition_t Decomposition type used in ITSOL
 *   (itsol_decomposition_t) and ZITSOL (zitsol_decomposition_t)
 *
 * @tparam z_itsol_clean_sparse_matrix Function used clean a
 *   z_itsol_sparse_matrix_t
 * @tparam z_itsol_clean_decomposition Function used to clean a
 *   z_itsol_decomposition_t
 * @tparam z_itsol_count_nnz_elements Function used to count the elements of a
 *   z_itsol_sparse_matrix_t
 * @tparam z_itsol_iluc_prepare_data_structure Function to prepare the data
 *   structure for ILUC preconditioner
 * @tparam z_itsol_iluc ILUC preconditioner function of ITSOL or ZITSOL
 *
 * @param sm The square sparse matrix that will be decomposed
 * @param lfil Number of preserved elements per row/column
 * @param tol Threshold for dropping
 * @param nargout Number of output arguments.
 */
template < typename octave_sparse_matrix_t, typename octave_column_vector_t,
  typename c_data_t, typename cpp_data_t, typename z_itsol_sparse_matrix_t,
  typename z_itsol_decomposition_t,
  int (*z_itsol_clean_sparse_matrix) (z_itsol_sparse_matrix_t *),
  int (*z_itsol_clean_decomposition) (z_itsol_decomposition_t *),
  int (*z_itsol_count_nnz_elements) (z_itsol_sparse_matrix_t *),
  int (*z_itsol_iluc_prepare_data_structure) (z_itsol_sparse_matrix_t *,
                          z_itsol_decomposition_t *, int),
  int (*z_itsol_iluc) (z_itsol_decomposition_t *, z_itsol_decomposition_t *,
               int, double, int, FILE *) >octave_value_list
generic_iluc (octave_sparse_matrix_t sm, int lfil, double tol, int drop,
          int nargout)
{
  octave_value_list retval;

  // Prepare Z_ITSOL data structures
  z_itsol_sparse_matrix_t *csmat =
    itsol_convert_to_cs < z_itsol_sparse_matrix_t, c_data_t,
    octave_sparse_matrix_t > (sm);
  z_itsol_decomposition_t *csmat_lu = new z_itsol_decomposition_t ();
  z_itsol_decomposition_t *lu = new z_itsol_decomposition_t ();

  // Convert to ITSOL formats
  int rsa = 2;            // 2: pattern symmetrization (add zeros)
  int ierr = z_itsol_iluc_prepare_data_structure (csmat, csmat_lu, rsa);
  if (ierr != 0)
    {
      error ("iluc: conversion error.");
      return retval;
    }

  // Z_ITSOL preconditioner requires an error outputstream. So redirect it to
  // an temporary file and read that file out in case of an error return code.
  FILE *redirected_stderr = gnulib::tmpfile ();
  if (!redirected_stderr)
    {
      error ("iluk: cannot open temporary file for error output.");
      return retval;
    }

  // Call ITSOL ILUC preconditioner
  ierr = z_itsol_iluc (csmat_lu, lu, lfil, tol, drop, redirected_stderr);
  if (ierr == -2)
    {
      error ("iluc: zero diagonal element found.");
      return retval;
    }
  else if (ierr == -1)
    {
      error ("iluc: Illegal value for lfil.");
      return retval;
    }
  else if (ierr != 0)
    {
      // Generate error message returned by Z_ITSOL
      std::rewind(redirected_stderr);
      std::stringstream err_msg_sstream;
      err_msg_sstream << "iluk: unknown internal error occured: ";
      char c = 0;

      // Read all printable characters from temporary file
      while (!std::feof (redirected_stderr)) {
        c = std::fgetc (redirected_stderr);
        if ((c < 32) || (c > 126))
          break;
        err_msg_sstream << c;
      }
      err_msg_sstream << std::endl;

      error (err_msg_sstream.str ().c_str ());
      return retval;
    }
  std::fclose (redirected_stderr);

  // Convert results to octave formats
  c_data_t *diag = lu->D;
  bool do_scale = false;
  bool add_diag = false;
  if (nargout == 2)
    add_diag = true;
  octave_sparse_matrix_t L =
    itsol_convert_lower_cs_to_octave < z_itsol_sparse_matrix_t, c_data_t,
    cpp_data_t, octave_sparse_matrix_t, z_itsol_count_nnz_elements > (lu->L,
                                      diag,
                                      do_scale,
                                      add_diag);
  octave_sparse_matrix_t U =
    itsol_convert_upper_cs_to_octave < z_itsol_sparse_matrix_t, c_data_t,
    cpp_data_t, octave_sparse_matrix_t, z_itsol_count_nnz_elements > (lu->U,
                                      diag,
                                      do_scale,
                                      add_diag);

  U = U.transpose ();        //FIXME: problem CSC - CRS check csmat_lu!!!

  retval (0) = octave_value (L);
  if (nargout == 2)
    retval (1) = octave_value (U);
  else if (nargout == 3)
    {
      retval (1) =
    octave_value (itsol_extract_diag < z_itsol_decomposition_t,
              octave_column_vector_t > (lu));
      retval (2) = octave_value (U);
    }

  // Tidy up Z_ITSOL structures
  z_itsol_clean_sparse_matrix (csmat);
  z_itsol_clean_decomposition (csmat_lu);
  z_itsol_clean_decomposition (lu);

  return retval;
}

DEFUN_DLD (iluc, args, nargout, "-*- texinfo -*-\n\
@deftypefn  {Loadable Function} {[@var{L}, @var{U}] =} iluc (@var{A})\n\
@deftypefnx {Loadable Function} {[@var{L}, @var{D}, @var{U}] =} iluc (@var{A})\n\
@deftypefnx {Loadable Function} {[...] =} iluc (@var{A}, @var{tol}, @var{lfil}, @var{drop})\n\
\n\
Computes the crout version incomplete LU-factorization (ILU) with threshold of \
@var{A}.\n\
\n\
NOTE: No pivoting is performed.\n\
\n\
@code{[@var{L}, @var{U}] = iluc (@var{A})} computes the default crout version\n\
ILU-factorization with threshold ILUT of @var{A}, such that \
@code{@var{L} * @var{U}} is an approximation of the square sparse matrix \
@var{A}.\n\
\n\
@code{[@var{L}, @var{D}, @var{U}] = iluc (@var{A})} computes the default crout \
version ILU-factorization with threshold ILUT of @var{A}, such that \
@code{(@var{L} + speye (length(@var{D}))) * (diag (@var{D}) + @var{U})}  is an \
approximation of the square sparse matrix @var{A}.\n\
\n\
@code{[...] = iluc (@var{A}, @var{tol}, @var{lfil}, @var{drop})} \
computes the ILU-factorization with threshold ILUT of @var{A}. Parameter \
@code{@var{tol}>=0} is the scalar double threshold. All elements \
@code{x<=@var{tol}} will be dropped in the factorization. Parameter \
@code{@var{lfil}>=0} is the scalar integer number of elements, that will be \
kept in the factorization. Thus only the largest @var{lfil}-elements in the \
i-th columns of @var{L} and @var{U} will be stored. Parameter @var{drop} \
specifies the dropping strategy:\n\n\
    0 : standard dropping strategy\n\n\
    1 : drops a term if corresponding perturbation relative to diagonal entry \
is small. For example for U:\n\n\
        @code{|| L(:,i) || * |u(i,j)/D(i,i)| < tol * | D(j,j) |}\n\n\
    2 : condition number estimator based on:\n\n\
        @code{rhs = (1, 1, ..., 1 )^T}\n\n\
    3 : condition number estimator based on:\n\n\
        @code{rhs = ((+/-)1, ..., (+/-)1 )^T + maximizing |v_k|}\n\n\
    4 : condition number estimator based on:\n\n\
        @code{rhs = ((+/-)1, ..., (+/-)1 )^T + maximizing || (v_(k+1), ..., v_n)^T ||}\n\n\
Default values are @code{@var{tol}=1e-4}, @code{@var{lfil}=50} and \
@code{@var{drop}=0}.\n\n\
NOTE: If @code{@var{tol}=0}, only the largest @var{lfil}-elements in each \
column of @var{L} and @var{U} will be kept.\n\
\n\
For more information about the algorithms themselves see:\n\
\n\
[1] Saad, Yousef: Iterative Methods for Sparse Linear Systems. Second Edition. \
Minneapolis, Minnesota: Siam 2003.\n\
\n\
This implementation makes use of the ITSOL library by Yousef Saad \
@email{saad@@cs.umn.edu} et al., University of Minnesota. (see\n\
@url{http://www-users.cs.umn.edu/@~{}saad/software/ITSOL/})\n\
@seealso{ilu, iluk, ilut, ilutp, ichol}\n\
@end deftypefn")
{
  octave_value_list retval;

#ifdef HAVE_ITSOL

  int nargin = args.length ();

  if (nargout < 2 || nargout > 3 || nargin < 1 || nargin > 4)
    {
      print_usage ();
      return retval;
    }

  double tol = 1e-4;
  if ((nargin >= 2) && (!args (1).is_scalar_type ()
            || ((tol = args (1).double_value ()) < 0)))
    error ("iluc: 2. parameter should be a positive double.");
  int lfil = 50;
  if ((nargin == 3) && (!args (2).is_scalar_type ()
            || ((lfil = args (2).int_value ()) < 0)))
    error ("iluc: 3. parameter should be a positive integer.");
  int drop = 0;
  if ((nargin == 4) && (!args (3).is_scalar_type ()
            || ((drop = args (3).int_value ()) < 0)
            || ((drop = args (3).int_value ()) > 4)))
    error ("iluc: 4. parameter should be a positive integer 0 <= drop <= 4.");

  if (!error_state)
    {
      if (!args (0).is_complex_type ())
        {
          // Prepare input matrix real case
          SparseMatrix sm = args (0).sparse_matrix_value ();
          retval = generic_iluc < SparseMatrix, ColumnVector,
            double, double, itsol_sparse_matrix_t,
            itsol_decomposition_t, itsol::cleanCS, itsol::cleanILU,
            itsol::nnz_cs, itsol::CSClumC, itsol::ilutc > (sm, lfil, tol,
                                   drop, nargout);
        }
          else
        {
          // Prepare input matrix complex case
          warning ("iluc: This function is experimental for complex input.");
          SparseComplexMatrix sm = args (0).sparse_complex_matrix_value ();
          retval = generic_iluc < SparseComplexMatrix, ComplexColumnVector,
            complex_conversion_t, Complex, zitsol_sparse_matrix_t,
            zitsol_decomposition_t, zitsol::zcleanCS, zitsol::zcleanILU,
            zitsol::znnzCS, zitsol::zCSClumC, zitsol::zilutc > (sm, lfil, tol,
                                   drop, nargout);
        }
    }

#else

  error ("iluc: ITSOL or ZITSOL not installed");

#endif

  return retval;
}

/* ITSOL test cases
%!shared n_tiny, n_small, n_medium, n_large, A_tiny, A_small, A_medium, A_large
%! n_tiny = 5;
%! n_small = 40;
%! n_medium = 600;
%! n_large = 10000;
%! A_tiny = spconvert([1 4 2 3 3 4 2 5; 1 1 2 3 4 4 5 5; 1 2 3 4 5 6 7 8]');
%! A_small = sprand(n_small, n_small, 1/n_small) + speye (n_small);
%! A_medium = sprand(n_medium, n_medium, 1/n_medium) + speye (n_medium);
%! A_large = sprand(n_large, n_large, 1/n_large/10) + speye (n_large);
%!testif HAVE_ITSOL && HAVE_ZITSOL
%! [L,U] = iluc ([]);
%! assert (isempty (L), logical (1));
%! assert (isempty (U), logical (1));
%! [L,D,U] = iluc ([]);
%! assert (isempty (L), logical (1));
%! assert (isempty (D), logical (1));
%! assert (isempty (U), logical (1));
%!error [L,U] = iluc (0);
%!error [L,D,U] = iluc (0);
%!error [L,U] = iluc (sparse (0));
%!error [L,D,U] = iluc (sparse (0));
%!testif HAVE_ITSOL && HAVE_ZITSOL
%! [L,U] = iluc (2);
%! assert (L, sparse (1));
%! assert (U, sparse (2));
%! [L,D,U] = iluc (2);
%! assert (L, sparse (0));
%! assert (D, 2);
%! assert (U, sparse (0));
%!testif HAVE_ITSOL && HAVE_ZITSOL
%! [L,U] = iluc (sparse (2));
%! assert (L, sparse (1));
%! assert (U, sparse (2));
%! [L,D,U] = iluc (sparse (2));
%! assert (L, sparse (0));
%! assert (D, 2);
%! assert (U, sparse (0));
%!testif HAVE_ITSOL && HAVE_ZITSOL
%! [L,U] = iluc (A_tiny);
%! assert (norm (A_tiny - L * U, "fro") / norm (A_tiny, "fro"), 0, n_tiny*eps);
%!testif HAVE_ITSOL && HAVE_ZITSOL
%! [L,U] = iluc (A_tiny, eps, n_tiny, 0);
%! assert (norm (A_tiny - L * U, "fro") / norm (A_tiny, "fro"), 0, n_tiny*eps);
%!testif HAVE_ITSOL && HAVE_ZITSOL
%! [L,D,U] = iluc (A_tiny);
%! assert (norm (A_tiny - (L + speye (length(D))) * (diag (D) + U), "fro") / norm (A_tiny, "fro"), 0, n_tiny*eps);
%!testif HAVE_ITSOL && HAVE_ZITSOL
%! [L,D,U] = iluc (A_tiny, eps, n_tiny, 0);
%! assert (norm (A_tiny - (L + speye (length(D))) * (diag (D) + U), "fro") / norm (A_tiny, "fro"), 0, n_tiny*eps);
%!testif HAVE_ITSOL && HAVE_ZITSOL
%! [L,U] = iluc (A_small);
%! assert (norm (A_small - L * U, "fro") / norm (A_small, "fro"), 0, 1);
%!testif HAVE_ITSOL && HAVE_ZITSOL
%! [L,U] = iluc (A_small, eps, n_small, 0);
%! assert (norm (A_small - L * U, "fro") / norm (A_small, "fro"), 0, 1);
%!testif HAVE_ITSOL && HAVE_ZITSOL
%! [L,D,U] = iluc (A_small);
%! assert (norm (A_small - (L + speye (length(D)) * (diag (D) + U)), "fro") / norm (A_small, "fro"), 0, 1);
%!testif HAVE_ITSOL && HAVE_ZITSOL
%! [L,D,U] = iluc (A_small, eps, n_small, 0);
%! assert (norm (A_small - (L + speye (length(D)) * (diag (D) + U)), "fro") / norm (A_small, "fro"), 0, 1);
%!testif HAVE_ITSOL && HAVE_ZITSOL
%! [L,U] = iluc (A_medium);
%! assert (norm (A_medium - L * U, "fro") / norm (A_medium, "fro"), 0, 1);
%!testif HAVE_ITSOL && HAVE_ZITSOL
%! [L,U] = iluc (A_medium, eps, n_medium, 0);
%! assert (norm (A_medium - L * U, "fro") / norm (A_medium, "fro"), 0, 1);
%!testif HAVE_ITSOL && HAVE_ZITSOL
%! [L,D,U] = iluc (A_medium);
%! assert (norm (A_medium - (L + speye (length(D)) * (diag (D) + U)), "fro") / norm (A_medium, "fro"), 0, 1);
%!testif HAVE_ITSOL && HAVE_ZITSOL
%! [L,D,U] = iluc (A_medium, eps, n_medium, 0);
%! assert (norm (A_medium - (L + speye (length(D)) * (diag (D) + U)), "fro") / norm (A_medium, "fro"), 0, 1);
%!testif HAVE_ITSOL && HAVE_ZITSOL
%! [L,U] = iluc (A_large);
%! assert (norm (A_large - L * U, "fro") / norm (A_large, "fro"), 0, 1);
%!testif HAVE_ITSOL && HAVE_ZITSOL
%! [L,U] = iluc (A_large, eps, n_large, 0);
%! assert (norm (A_large - L * U, "fro") / norm (A_large, "fro"), 0, 1);
%!testif HAVE_ITSOL && HAVE_ZITSOL
%! [L,D,U] = iluc (A_large);
%! assert (norm (A_large - (L + speye (length(D)) * (diag (D) + U)), "fro") / norm (A_large, "fro"), 0, 1);
%!testif HAVE_ITSOL && HAVE_ZITSOL
%! [L,D,U] = iluc (A_large, eps, n_large, 0);
%! assert (norm (A_large - (L + speye (length(D)) * (diag (D) + U)), "fro") / norm (A_large, "fro"), 0, 1);
*/

/* ZITSOL test cases
%!shared n_tiny, n_small, n_medium, n_large, A_tiny, A_small, A_medium, A_large
%! n_tiny = 5;
%! n_small = 40;
%! n_medium = 600;
%! n_large = 10000;
%! A_tiny = spconvert([1 4 2 3 3 4 2 5; 1 1 2 3 4 4 5 5; 1 2 3 4 5 6 7 8]');
%! A_tiny(1,1) += 1i;
%! A_small = sprand(n_small, n_small, 1/n_small) + i * sprand(n_small, n_small, 1/n_small) + speye (n_small);
%! A_medium = sprand(n_medium, n_medium, 1/n_medium) + i * sprand(n_medium, n_medium, 1/n_medium) + speye (n_medium);
%! A_large = sprand(n_large, n_large, 1/n_large/10) + i * sprand(n_large, n_large, 1/n_large/10) + speye (n_large);
%!testif HAVE_ITSOL && HAVE_ZITSOL
%! [L,U] = iluc ([]);
%! assert (isempty (L), logical (1));
%! assert (isempty (U), logical (1));
%! [L,D,U] = iluc ([]);
%! assert (isempty (L), logical (1));
%! assert (isempty (D), logical (1));
%! assert (isempty (U), logical (1));
%!error [L,U] = iluc (0+0i);
%!error [L,D,U] = iluc (0+0i);
%!error [L,U] = iluc (0i);
%!error [L,D,U] = iluc (0i);
%!error [L,U] = iluc (sparse (0+0i));
%!error [L,D,U] = iluc (sparse (0+0i));
%!error [L,U] = iluc (sparse (0i));
%!error [L,D,U] = iluc (sparse (0i));
%!testif HAVE_ITSOL && HAVE_ZITSOL
%! [L,U] = iluc (2+0i);
%! assert (L, sparse (1));
%! assert (U, sparse (2));
%! [L,D,U] = iluc (2+0i);
%! assert (L, sparse (0));
%! assert (D, 2);
%! assert (U, sparse (0));
%! [L,U] = iluc (2+2i);
%! assert (L, sparse (1));
%! assert (U, sparse (2+2i));
%! [L,D,U] = iluc (2+2i);
%! assert (L, sparse (0));
%! assert (D, 2+2i);
%! assert (U, sparse (0));
%! [L,U] = iluc (2i);
%! assert (L, sparse (1));
%!# assert (U, sparse (2i));
%! [L,D,U] = iluc (2i);
%! assert (L, sparse (0));
%! assert (D, 2i);
%! assert (U, sparse (0));
%!testif HAVE_ITSOL && HAVE_ZITSOL
%! [L,U] = iluc (sparse (2+0i));
%! assert (L, sparse (1));
%! assert (U, sparse (2));
%! [L,D,U] = iluc (sparse (2+0i));
%! assert (L, sparse (0));
%! assert (D, 2);
%! assert (U, sparse (0));
%! [L,U] = iluc (sparse (2+2i));
%! assert (L, sparse (1));
%! assert (U, sparse (2+2i));
%! [L,D,U] = iluc (sparse (2+2i));
%! assert (L, sparse (0));
%! assert (D, 2+2i);
%! assert (U, sparse (0));
%! [L,U] = iluc (sparse (2i));
%! assert (L, sparse (1));
%!# assert (U, sparse (2i));
%! [L,D,U] = iluc (sparse (2i));
%! assert (L, sparse (0));
%! assert (D, 2i);
%! assert (U, sparse (0));
%!testif HAVE_ITSOL && HAVE_ZITSOL
%! [L,U] = iluc (A_tiny);
%! assert (norm (A_tiny - L * U, "fro") / norm (A_tiny, "fro"), 0, n_tiny*eps);
%!testif HAVE_ITSOL && HAVE_ZITSOL
%! [L,U] = iluc (A_tiny, eps, n_tiny, 0);
%! assert (norm (A_tiny - L * U, "fro") / norm (A_tiny, "fro"), 0, n_tiny*eps);
%!testif HAVE_ITSOL && HAVE_ZITSOL
%! [L,D,U] = iluc (A_tiny);
%! assert (norm (A_tiny - (L + speye (length(D))) * (diag (D) + U), "fro") / norm (A_tiny, "fro"), 0, n_tiny*eps);
%!testif HAVE_ITSOL && HAVE_ZITSOL
%! [L,D,U] = iluc (A_tiny, eps, n_tiny, 0);
%! assert (norm (A_tiny - (L + speye (length(D))) * (diag (D) + U), "fro") / norm (A_tiny, "fro"), 0, n_tiny*eps);
%!testif HAVE_ITSOL && HAVE_ZITSOL
%! [L,U] = iluc (A_small);
%! assert (norm (A_small - L * U, "fro") / norm (A_small, "fro"), 0, 1);
%!testif HAVE_ITSOL && HAVE_ZITSOL
%! [L,U] = iluc (A_small, eps, n_small, 0);
%! assert (norm (A_small - L * U, "fro") / norm (A_small, "fro"), 0, 1);
%!testif HAVE_ITSOL && HAVE_ZITSOL
%! [L,D,U] = iluc (A_small);
%! assert (norm (A_small - (L + speye (length(D)) * (diag (D) + U)), "fro") / norm (A_small, "fro"), 0, 1);
%!testif HAVE_ITSOL && HAVE_ZITSOL
%! [L,D,U] = iluc (A_small, eps, n_small, 0);
%! assert (norm (A_small - (L + speye (length(D)) * (diag (D) + U)), "fro") / norm (A_small, "fro"), 0, 1);
%!testif HAVE_ITSOL && HAVE_ZITSOL
%! [L,U] = iluc (A_medium);
%! assert (norm (A_medium - L * U, "fro") / norm (A_medium, "fro"), 0, 1);
%!testif HAVE_ITSOL && HAVE_ZITSOL
%! [L,U] = iluc (A_medium, eps, n_medium, 0);
%! assert (norm (A_medium - L * U, "fro") / norm (A_medium, "fro"), 0, 1);
%!testif HAVE_ITSOL && HAVE_ZITSOL
%! [L,D,U] = iluc (A_medium);
%! assert (norm (A_medium - (L + speye (length(D)) * (diag (D) + U)), "fro") / norm (A_medium, "fro"), 0, 1);
%!testif HAVE_ITSOL && HAVE_ZITSOL
%! [L,D,U] = iluc (A_medium, eps, n_medium, 0);
%! assert (norm (A_medium - (L + speye (length(D)) * (diag (D) + U)), "fro") / norm (A_medium, "fro"), 0, 1);
%!testif HAVE_ITSOL && HAVE_ZITSOL
%! [L,U] = iluc (A_large);
%! assert (norm (A_large - L * U, "fro") / norm (A_large, "fro"), 0, 1);
%!testif HAVE_ITSOL && HAVE_ZITSOL
%! [L,U] = iluc (A_large, eps, n_large, 0);
%! assert (norm (A_large - L * U, "fro") / norm (A_large, "fro"), 0, 1);
%!testif HAVE_ITSOL && HAVE_ZITSOL
%! [L,D,U] = iluc (A_large);
%! assert (norm (A_large - (L + speye (length(D)) * (diag (D) + U)), "fro") / norm (A_large, "fro"), 0, 1);
%!testif HAVE_ITSOL && HAVE_ZITSOL
%! [L,D,U] = iluc (A_large, eps, n_large, 0);
%! assert (norm (A_large - (L + speye (length(D)) * (diag (D) + U)), "fro") / norm (A_large, "fro"), 0, 1);
*/
