/**
 * Copyright (C) 2013 Kai T. Ohlhus <k.ohlhus@gmail.com>
 *
 * This file is part of Octave.
 *
 * Octave is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * Octave is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Octave; see the file COPYING.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * This is the octave interface to itsol. For more information see
 * http://www-users.cs.umn.edu/~saad/software/ITSOL/
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "defun-dld.h"

// Auxiliary ITSOL libraries
#include "itsol_util.h"

/**
 * Type generic ILUTP implementation that performs the same procedure for real
 * and complex input and calls ITSOL or ZITSOL library functions accordingly.
 *
 * @tparam octave_sparse_matrix_t The type of the Octave input matrix. Intended
 *   values for ITSOL (SparseMatrix) and for ZITSOL (SparseComplexMatrix)
 * @tparam octave_column_vector_t The type for the diagonal vector that will be
 *   returned to Octave. Intended values for ITSOL (ColumnVector) and for ZITSOL
 *   (ComplexColumnVector)
 * @tparam c_data_t Data type used in ITSOL (double) and ZITSOL
 *   (complex_conversion_t)
 * @tparam cpp_data_t Data type used in Octave. Indended values for ITSOL
 *   (double) and for ZITSOL (Complex)
 * @tparam z_itsol_sparse_matrix_t Sparse matrix storage type used in ITSOL
 *   (itsol_sparse_matrix_t) and ZITSOL (zitsol_sparse_matrix_t)
 * @tparam z_itsol_z_ilutp_decomposition_t Decomposition type used in ITSOL
 *   (itsol_ilutp_decomposition_t) and ZITSOL (zitsol_zilutp_decomposition_t)
 *
 * @tparam z_itsol_setup_z_ilutp_decomposition Function used to initialize a
 *   z_itsol_z_ilut_decomposition_t
 * @tparam z_itsol_clean_sparse_matrix Function used clean a
 *   z_itsol_sparse_matrix_t
 * @tparam z_itsol_clean_z_ilutp_decomposition Function used to clean a
 *   z_itsol_z_ilut_decomposition_t
 * @tparam z_itsol_count_nnz_elements Function used to count the elements of a
 *   z_itsol_sparse_matrix_t
 * @tparam z_itsol_ilutp ILUTP preconditioner function of ITSOL or ZITSOL
 *
 * @param sm The square sparse matrix that will be decomposed
 * @param tol tolerance threshold
 * @param lfil level of fill
 * @param permtol permutation tolerance
 * @param mband band structure
 * @param nargout Number of output arguments.
 */
template < typename octave_sparse_matrix_t, typename octave_column_vector_t,
  typename c_data_t, typename cpp_data_t, typename z_itsol_sparse_matrix_t,
  typename z_itsol_z_ilutp_decomposition_t,
  int (*z_itsol_setup_z_ilutp_decomposition)
    (z_itsol_z_ilutp_decomposition_t *, int),
  int (*z_itsol_clean_sparse_matrix) (z_itsol_sparse_matrix_t *),
  int (*z_itsol_clean_z_ilutp_decomposition)
    (z_itsol_z_ilutp_decomposition_t *, int),
  int (*z_itsol_count_nnz_elements) (z_itsol_sparse_matrix_t *),
  int (*z_itsol_ilutp) (z_itsol_sparse_matrix_t *, double *, int *,
    double, int, z_itsol_z_ilutp_decomposition_t *) >octave_value_list
generic_ilutp (octave_sparse_matrix_t sm, double *tol, int *lfil,
  double permtol, int mband, int nargout)
{
  octave_value_list retval;
  
  int n = sm.cols (); // assume square matrix

  // Prepare ITSOL data structures
  z_itsol_sparse_matrix_t *csmat =
    itsol_convert_to_cs < z_itsol_sparse_matrix_t, c_data_t,
    octave_sparse_matrix_t > (sm);
  z_itsol_z_ilutp_decomposition_t *lup = new z_itsol_z_ilutp_decomposition_t ();
  int ierr = z_itsol_setup_z_ilutp_decomposition (lup, n);
  if (ierr != 0)
    {
      error("ilutp: memory allocation error.");
      return retval;
    }
  lup->perm2 = new int[n]();

  // Call ITSOL ILUT preconditioner
  ierr = z_itsol_ilutp (csmat, tol, lfil, permtol, mband, lup); //z_itsol_ilutp
  if (ierr == 6)
    {
      error ("ilutp: zero row encountered.");
      return retval;
    }
  else if (ierr == 5)
    {
      error ("ilutp: illegal value for lfil.");
      return retval;
    }
  else if (ierr == 2)
    {
      error ("ilutp: memory allocation error.");
      return retval;
    }
  else if (ierr != 0)
    {
      error ("ilutp: unknown internal error occured.");
      return retval;
    }

  // Convert results to octave formats
  c_data_t *diag = NULL;
  bool do_scale = false;
  bool add_diag = false;
  octave_sparse_matrix_t L =
    itsol_convert_lower_cs_to_octave < z_itsol_sparse_matrix_t, c_data_t,
    cpp_data_t, octave_sparse_matrix_t, z_itsol_count_nnz_elements > (lup->U,
                                      diag,
                                      do_scale,
                                      add_diag);
  octave_sparse_matrix_t U =
    itsol_convert_upper_cs_to_octave < z_itsol_sparse_matrix_t, c_data_t,
    cpp_data_t, octave_sparse_matrix_t, z_itsol_count_nnz_elements > (lup->L,
                                      diag,
                                      do_scale,
                                      add_diag);

  retval (0) = octave_value (L);
  retval (1) = octave_value (U);
  if (nargout == 3)
    {
      retval (2) = 
        octave_value (itsol_extract_perm2 <z_itsol_z_ilutp_decomposition_t, 
          octave_column_vector_t>(lup));
    }

  // Tidy up Z_ITSOL structures
  z_itsol_clean_sparse_matrix (csmat);
  z_itsol_clean_z_ilutp_decomposition (lup, 0); //FIXME: clarify last parameter

  return retval;
}

DEFUN_DLD (ilutp, args, nargout,
"-*- texinfo -*-\n\
@deftypefn  {Loadable Function} {[@var{L}, @var{U}, @var{P}] =} ilutp (@var{A})\n\
@deftypefnx {Loadable Function} {[@var{L}, @var{U}, @var{P}] =} ilutp (@var{A}, @var{tol}, @var{lfil}, @var{droptol}, @var{mband})\n\
\n\
Computes the incomplete LU-factorization (ILU) with threshold and column\n\
pivoting of @var{A}.\n\
\n\
No implementation so far. \
\n\
For more information about the algorithms themselves see:\n\n\
[1] Saad, Yousef: Iterative Methods for Sparse Linear Systems. Second Edition. \
Minneapolis, Minnesota: Siam 2003.\n\
\n\
This implementation makes use of the ITSOL library by Yousef Saad \
@email{saad@@cs.umn.edu} et al., University of Minnesota. (see\n\
@url{http://www-users.cs.umn.edu/@~{}saad/software/ITSOL/})\n\
@seealso{ilu, iluk, ilut, iluc, ichol}\n\
@end deftypefn")
{
  octave_value_list retval;
  error ("ilutp: Not yet implemented."); // FIXME
  return retval;

#if defined (HAVE_ITSOL) && defined (HAVE_ZITSOL)

  int nargin = args.length ();

  if (nargout < 2 || nargout > 3  || nargin < 1 || nargin > 5)
    {
      print_usage ();
      return retval;
    }
    
  OCTAVE_LOCAL_BUFFER(double, tol, 7);
  tol[5] = 1e-8;
  tol[6] = 1e-8;
  if (nargin >= 2)
    {
      bool bad_input = false;
      if (args (1).is_scalar_type ()
                && ((tol[5] = tol[6] = args (1).double_value ()) < 0))
        bad_input = true;
      else if (!args (1).is_scalar_type ()) //FIXME Test for matrix input
        {
          Array<double> tol_arr = args (1).vector_value();
          if ((tol_arr.length () != 2) || ((tol[5] = tol_arr (0)) < 0)
                    || ((tol[6] = tol_arr (1)) < 0))
            {
              bad_input = true;
            }
        }
      if (bad_input)
        error ("ilutp: 2. parameter should be a positive double or a vector of two positive doubles.");
    }
  OCTAVE_LOCAL_BUFFER(int, lfil, 7);
  lfil[5] = 50;
  lfil[6] = 50;
  if (nargin >= 3)
    {
      bool bad_input = false;
      if (args (2).is_scalar_type ()
                && ((lfil[5] = lfil[6] = args (2).int_value ()) < 0))
        bad_input = true;
      else if (!args (1).is_scalar_type ()) //FIXME Test for matrix input
        {
          Array<int> lfil_arr = args (2).int_vector_value ();
          if ((lfil_arr.length () != 2) || ((lfil[5] = lfil_arr (0)) < 0)
                    || ((lfil[6] = lfil_arr (1)) < 0))
            {
              bad_input = true;
            }
        }
      if (bad_input)
        {
          error ("ilutp: 3. parameter should be a positive integer or a vector of two positive integers.");
        }
    }
  double permtol = 1e-8;
  if ((nargin >= 4) && (!args (3).is_scalar_type ()
            || ((permtol = args (3).double_value ()) < 0)))
    error ("ilut: 4. parameter should be a positive double.");
  int mband = 5;
  if ((nargin == 5) && (!args (4).is_scalar_type ()
            || ((mband = args (4).int_value ()) < 0)))
    error ("ilut: 5. parameter should be a positive integer.");

  if (!error_state)
    {
      if (!args (0).is_complex_type ())
        {
          // Prepare input matrix real case
          SparseMatrix sm = args (0).sparse_matrix_value ();
          retval = generic_ilutp < SparseMatrix, ColumnVector,
            double, double, itsol_sparse_matrix_t,
            itsol_ilutp_decomposition_t, itsol::setupILUT, itsol::cleanCS,
            itsol::cleanILUT, itsol::nnz_cs, itsol::ilutpC >
              (sm, tol, lfil, permtol, mband, nargout);
        }
      else
        {
          // Prepare input matrix complex case
          SparseComplexMatrix sm = args (0).sparse_complex_matrix_value ();
          retval = generic_ilutp < SparseComplexMatrix, ComplexColumnVector,
            complex_conversion_t, Complex, zitsol_sparse_matrix_t,
            zitsol_zilutp_decomposition_t, zitsol::zsetupILUT,
            zitsol::zcleanCS, zitsol::zcleanILUT, zitsol::znnzCS,
            zitsol::zilutpC > 
              (sm, tol, lfil, permtol, mband, nargout);
        }
    }

#else

  error ("ilutp: ITSOL or ZITSOL not installed");

#endif

  return retval;
}

