/**
 * Copyright (C) 2013      Kai T. Ohlhus <k.ohlhus@gmail.com>
 * Copyright (C) 2012-2013 Wei Jin       <3162073@gmail.com>
 *
 * This file is part of Octave.
 *
 * Octave is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * Octave is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Octave; see the file COPYING.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * This is the octave interface to itsol. For more information see
 * http://www-users.cs.umn.edu/~saad/software/ITSOL/
 */

// ITSOL libraries
namespace itsol
{
  extern "C"
  {
#include <itsol/globheads.h>
#include <itsol/protos.h>
#include <itsol/defs.h>
#include <itsol/ios.h>
  }
}

// ZITSOL libraries
namespace zitsol
{
  extern "C"
  {
#include <zitsol/zheads.h>
#include <zitsol/zprotos.h>
#include <zitsol/zdefs.h>
#include <zitsol/ios.h>
  }
}

// Define consistent identifiers for ITSOL and ZITSOL types and functions
typedef
  double _Complex
  complex_conversion_t;
typedef
  itsol::SparMat
  itsol_sparse_matrix_t;
typedef
  itsol::ILUSpar
  itsol_decomposition_t;
typedef
  itsol::IluSpar
  itsol_ilutp_decomposition_t;
typedef
  zitsol::zSparMat
  zitsol_sparse_matrix_t;
typedef
  zitsol::zILUSpar
  zitsol_decomposition_t;
typedef
  zitsol::zIluSpar
  zitsol_zilutp_decomposition_t;

/**
 * Some Information about the sparse matrix storage formats used here:
 * 
 * 0.) Naming conventions
 *     
 *     n   = dimension of sqaure sparse matrix
 *     nnz = number of non-zero elements in the sqaure sparse matrix
 *
 * 1.) Octave compressed sparse column (CSC) (a.k.a. compressed column storage (CCS))
 *     http://arxiv.org/abs/cs.MS/0604006
 * 
 *     cidx (column vector) size: n + 1
 *     ridx (row vector)    size: nnz
 *     data (data vector)   size: nnz
 *
 * 2.) Coordinate format (COO)
 *
 *     coo_c (column vector) size: nnz
 *     coo_r (row vector)    size: nnz
 *     coo_d (data vector)   size: nnz
 *
 * 3.) ITSOL format (cs) row storage (a.k.a. compressed row storage (CRS))
 *
 *     SpaFmt->n                           size: 1
 *     SpaFmt->nzcount (lengths of rows)   size: n
 *     SpaFmt->*ja     (column vector)     size: n   ja[i]-size: nzcount[i]
 *     SpaFmt->*ma     (data vector)       size: n   ma[i]-size: nzcount[i]
 */

/**
 * Converts Octave (CSC) format to (Z)ITSOL (cs) format.
 * NOTE: implicit transposition of input matrix.
 *
 * @tparam z_itsol_sparse_matrix_t appropiate for ITSOL (real matrix) is "SparMat" and for
 *   ZITSOL (complex matrix) it is "zSparMat".
 * @tparam c_data_t appropiate for ITSOL (real matrix) is "double" and for
 *   ZITSOL (complex matrix) it is "complex_conversion_t".
 * @tparam octave_sparse_matrix_t appropiate for ITSOL (real matrix) is "SparseMatrix" and
 *   for ZITSOL (complex matrix) it is "ComplexSparseMatrix".
 * @param sm Octave (CSC) format sparse square matrix
 * @return Pointer to (Z)ITSOL (cs) (z)SparMat-struct
 */
template < typename z_itsol_sparse_matrix_t, typename c_data_t,
  typename octave_sparse_matrix_t >
  z_itsol_sparse_matrix_t *
itsol_convert_to_cs (const octave_sparse_matrix_t & sm)
{
  // Get matrix dimension (assume square)
  int n = sm.cols ();
  // Pointer to Octave matrix
  int * cidx = sm.cidx ();
  int * ridx = sm.ridx ();
  c_data_t * data = reinterpret_cast < c_data_t * >(sm.data ());

  // Get memory for (Z)ITSOL (cs) matrix
  z_itsol_sparse_matrix_t *
    csmat = new z_itsol_sparse_matrix_t ();    // basic structure
  csmat->nzcount = new int[n] ();    // row vector
  csmat->ja = new int *[n] ();    // column vector
  csmat->ma = new c_data_t *[n] ();    // data vector

  // Convert Octave (CSC) format to (Z)ITSOL (cs) format
  // Note: here the implicit transposition happens
  csmat->n = n;
  for (int i = 0; i < n; i++)
    {
      int number_of_nnz = cidx[i + 1] - cidx[i];
      csmat->nzcount[i] = number_of_nnz;
      csmat->ja[i] = new int[number_of_nnz] ();
      csmat->ma[i] = new c_data_t[number_of_nnz] ();
      for (int j = 0; j < number_of_nnz; j++)
        {
          csmat->ja[i][j] = *(ridx++);
          csmat->ma[i][j] = *(data++);
        }
    }

  return csmat;
}

/**
 * Converts ITSOL (cs) format upper triangular matrix to Octave (CSC) format.
 *
 * Do scale:
 *
 *      [ 0   x   x          x ]   [ 1/d_0                           ]
 *      [     0   x          x ]   [       1/d_1                     ]
 * sm = [         0          x ] * [             1/d_2               ]
 *      [             ...    x ]   [                   ...           ]
 *      [                    0 ]   [                       1/d_(n-1) ]
 *
 * sm =             cs           *             diag(1./scale)
 *
 * Add diagonal:
 *
 *      [ 1/d_0   x     x            x      ]
 *      [       1/d_1   x            x      ]
 * sm = [             1/d_2          x      ]
 *      [                    ...     x      ]
 *      [                         1/d_(n-1) ]
 * 
 * @param cs pointer to a ITSOL (cs) format SparMat-struct (e.g. result of an
 *        ILU-decomposition).
 * @param diag diagonal vector of same size as triangular matrix. May not be
 *        NULL if used in combination with the booleans do_scale or add_diag.
 * @param do_scale if true it performs the operation explained above.
 * @param add_diag if true it performs the operation explained above. Otherwise
 *        the diagonal remains as it is.
 * @return Octave (CSC) format matrix
 */
template < typename z_itsol_sparse_matrix_t, typename c_data_t,
  typename cpp_data_t, typename octave_sparse_matrix_t,
  int (*z_itsol_count_nnz_elements) (z_itsol_sparse_matrix_t *) >
  octave_sparse_matrix_t
itsol_convert_upper_cs_to_octave (z_itsol_sparse_matrix_t * cs,
                  c_data_t * diag, bool do_scale,
                  bool add_diag)
{
  // Get matrix properties (assume square)
  int n = cs->n;
  int nnz = z_itsol_count_nnz_elements (cs);
  int diag_size = 0;
  if (add_diag)
    {
      nnz += n;
      diag_size = 1;
    }

  // Get memory for Octave (CSC) matrix
  octave_sparse_matrix_t
  sm (n, n, nnz);

  // Get column indices.
  sm.cidx (0) = 0;
  for (int i = 0; i < n; i++)
    sm.cidx (i + 1) = sm.cidx (i) + (cs->nzcount[i]) + diag_size;

  int k = 0;
  cpp_data_t d;
  c_data_t * d_ptr = reinterpret_cast < c_data_t * >(&d);
  for (int i = 0; i < n; i++)
    {
      for (int j = 0; j < (cs->nzcount[i]); j++)
        {
          sm.ridx (k) = cs->ja[i][j];
          sm.data (k) = cs->ma[i][j];
          if (do_scale)
            {
              *d_ptr = diag[sm.ridx (k)];
              sm.data (k) /= d;
            }
          k++;
        }
      if (add_diag)
        {
          sm.ridx (k) = i;
          sm.data (k) = 1.0 / diag[i];
          k++;
        }
    }

  return sm;
}

/**
 * Converts ITSOL (cs) format lower triangular matrix to Octave (CSC) format.
 *
 * Scaling:
 *
 *      [ scale[0]                         ]   [ 0   x   x          x ]
 *      [          scale[1]                ]   [     0   x          x ]
 * sm = [                   ...            ] * [         0          x ]
 *      [                                  ]   [             ...    x ]
 *      [                       scale[n-1] ]   [                    0 ]
 *
 * sm =             diag(scale)              *             cs
 *
 *
 * Adding diagonal:
 *
 *      [ 1   x   x          x ]
 *      [     1   x          x ]
 * sm = [         1          x ]
 *      [             ...    x ]
 *      [                    1 ]
 * 
 * @param cs pointer to a ITSOL (cs) format SparMat-struct (e.g. result of an
 *        ILU-decomposition).
 * @param diag diagonal vector of same size as triangular matrix. May not be
 *        NULL if used in combination with the booleans do_scale or add_diag.
 * @param do_scale if true it performs the operation explained above.
 * @param add_diag if true it performs the operation explained above. Otherwise
 *        the diagonal remains as it is.
 * @return Octave (CSC) format matrix
 */
template < typename z_itsol_sparse_matrix_t, typename c_data_t,
  typename cpp_data_t, typename octave_sparse_matrix_t,
  int (*z_itsol_count_nnz_elements) (z_itsol_sparse_matrix_t *) >
  octave_sparse_matrix_t
itsol_convert_lower_cs_to_octave (z_itsol_sparse_matrix_t * cs,
                  c_data_t * diag, bool do_scale,
                  bool add_diag)
{
  // Get matrix properties (assume square)
  int n = cs->n;
  int nnz = z_itsol_count_nnz_elements (cs);
  int diag_size = 0;
  if (add_diag)
    {
      nnz += n;
      diag_size = 1;
    }

  // Get memory for Octave (CSC) matrix
  octave_sparse_matrix_t
  sm (n, n, nnz);

  // Get column indices.
  sm.cidx (0) = 0;
  for (int i = 0; i < n; i++)
    sm.cidx (i + 1) = sm.cidx (i) + (cs->nzcount[i]) + diag_size;

  int k = 0;
  cpp_data_t d;
  c_data_t * d_ptr = reinterpret_cast < c_data_t * >(&d);
  for (int i = 0; i < n; i++)
    {
      if (add_diag)
        {
          sm.ridx (k) = i;
          sm.data (k) = 1.0;
          k++;
        }
      for (int j = 0; j < (cs->nzcount[i]); j++)
        {
          sm.ridx (k) = cs->ja[i][j];
          sm.data (k) = cs->ma[i][j];
          if (do_scale)
            {
              *d_ptr = diag[i];
              sm.data (k) *= d;
            }
          k++;
        }
    }

  return sm;
}

/**
 * Extracts the diagonal vector from an (Z)ITSOL (z)ILUSpar-struct.
 *
 * @tparam z_itsol_sparse_matrix_t appropiate for ITSOL (real matrix) is "ILUSpar" and for
 *   ZITSOL (complex matrix) it is "zILUSpar".
 * @tparam octave_column_vector_t appropiate for ITSOL (real matrix) is "ColumnVector" and
 *   for ZITSOL (complex matrix) it is "ComplexColumnVector".
 * @param lu pointer to an (Z)ITSOL (z)ILUSpar-struct
 * @return the diagonal vector as (Complex)ColumnVector
 */
template < typename z_itsol_sparse_matrix_t,
  typename octave_column_vector_t > octave_column_vector_t
itsol_extract_diag (z_itsol_sparse_matrix_t * lu)
{
  octave_column_vector_t
  D (lu->n);
  for (int i = 0; i < lu->n; i++)
    D (i) = 1 / (lu->D[i]);
  return D;
}

/**
 * Extracts the permutation vector from an ITSOL IluSpar-struct.
 *
 * @tparam z_itsol_sparse_matrix_t appropiate for ITSOL (real matrix) is "IluSpar" and for
 *   ZITSOL (complex matrix) it is "zIluSpar".
 * @tparam octave_column_vector_t appropiate for ITSOL (real matrix) is "ColumnVector" and
 *   for ZITSOL (complex matrix) it is "ComplexColumnVector".
 * @param lu pointer to an ITSOL IluSpar-struct
 * @return the permutation vector as (Complex)ColumnVector
 */
template < typename z_itsol_sparse_matrix_t,
  typename octave_column_vector_t > octave_column_vector_t
itsol_extract_perm2 (z_itsol_sparse_matrix_t * lu)
{
  octave_column_vector_t
  D (lu->n);
  for (int i = 0; i < lu->n; i++)
    D (i) = lu->perm2[i];
  return D;
}

